#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Clusterer.
"""

import numpy as np
from numbers import Number

from ..common import Model
from ..common import UnsupervisedTrainer
from ..common import Proxy
from ..common import Utils
from ..common import LearningEnvironmentBuilder

from ..core import Cache
from ..common import gateway

class ClusteringTrainer(UnsupervisedTrainer, Proxy):
    """Clustering trainer.
    """
    def __init__(self, proxy):
        """Constructs a new instance of ClusteringTrainer.
        """
        Proxy.__init__(self, proxy)

    def fit(self, X):
        if isinstance(X, Cache):
            return self.fit_on_cache(X)

        X_java = Utils.to_java_double_array(X)
        y_java = Utils.to_java_double_array(np.zeros(X.shape[0]))

        java_model = self.proxy.fit(X_java, y_java, None)

        return Model(java_model, False)
    
    def fit_on_cache(self, cache):
        if not isinstance(cache, Cache):
            raise Exception("Unexpected type of cache (%s)." % type(cache))

        java_model = self.proxy.fitOnCache(cache.proxy, cache.cache_filter, Proxy.proxy_or_none(cache.preprocessor))
        return Model(java_model, False)

class GMMClusteringTrainer(ClusteringTrainer):
    """GMM clustring trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(), eps=1e-3, count_of_components=2,
                 max_iter=10, max_count_of_init_tries=3, max_count_of_clusters=2, max_likelihood_divirgence=5.0,
                 min_elements_for_new_cluster=300, min_cluster_probability=0.05):
        """Constructs a new instance of GMM clustring trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        count_of_components : Count of components.
        max_iter : Max number of iterations.
        max_count_of_init_tries : Max count of init tries.
        max_count_of_clusters : Max count of clusters.
        max_likelihood_divirgence : Max likelihood divirgence.
        min_elements_for_new_cluster : Min elements for new cluster.
        min_cluster_probability : Min cluster probability.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.clustering.gmm.GmmTrainer()
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))
        proxy.withInitialCountOfComponents(count_of_components)
        proxy.withMaxCountIterations(max_iter)
        proxy.withEps(eps)
        proxy.withMaxCountOfInitTries(max_count_of_init_tries)
        proxy.withMaxCountOfClusters(max_count_of_clusters)
        proxy.withMaxLikelihoodDivergence(max_likelihood_divirgence)
        proxy.withMinElementsForNewCluster(min_elements_for_new_cluster)
        proxy.withMinClusterProbability(min_cluster_probability)

        ClusteringTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class KMeansClusteringTrainer(ClusteringTrainer):
    """KMeans clustring trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(), amount_of_clusters=2,
                 max_iter=10, eps=1e-4, distance='euclidean'):
        """Constructs a new instance of KMeans clustering trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        amount_of_clusters : Amount of clusters.
        max_iter : Max number of iterations.
        eps : Epsilon.
        distance : Distance measure ('euclidean', 'hamming', 'manhattan').
        """
        proxy = gateway.jvm.org.apache.ignite.ml.clustering.kmeans.KMeansTrainer()
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))
        proxy.withAmountOfClusters(amount_of_clusters)
        proxy.withMaxIterations(max_iter)
        proxy.withEpsilon(eps)

        java_distance = None
        if distance == 'euclidean':
            java_distance = gateway.jvm.org.apache.ignite.ml.math.distances.EuclideanDistance()
        elif distance == 'hamming':
            java_distance = gateway.jvm.org.apache.ignite.ml.math.distances.HammingDistance()
        elif distance == 'manhattan':
            java_distance = gateway.jvm.org.apache.ignite.ml.math.distances.ManhattanDistance()
        elif distance:
            raise Exception("Unknown distance type : %s" % distance)
        proxy.withDistance(java_distance)

        ClusteringTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Ignite inference functionality.
"""

from ..common import Proxy
from ..common import Utils
from ..common import Model
from ..common import gateway
from copy import copy

class DistributedModel(Model):
    def __init__(self, ignite, reader, parser, instances=1, max_per_node=1):
        """Constructs a new instance of distributed model.

        Parameters
        ----------
        ignite : Ignite instance.
        reader : Model reader.
        parser : Model parser.
        mdl : Model.
        instances : Number of worker instances.
        max_per_node : Max number of worker per node.
        """
        super(DistributedModel, self).__init__(None, False)

        self.ignite = ignite
        self.reader = reader
        self.parser = parser
        self.instances = instances
        self.max_per_node = max_per_node

    def __enter__(self):
        self.proxy = [gateway.jvm.org.apache.ignite.ml.inference.builder.IgniteDistributedModelBuilder(
            self.ignite.ignite,
            self.instances,
            self.max_per_node
        ).build(r, self.parser) for r in self.reader]
        return self

    def __exit__(self, t, v, trace):
        if self.proxy is not None:
            for p in self.proxy:
                p.close()
            self.proxy = None
        return False

class XGBoostDistributedModel(DistributedModel):
    def __init__(self, ignite, mdl, instances=1, max_per_node=1):
        reader = [gateway.jvm.org.apache.ignite.ml.inference.reader.FileSystemModelReader(mdl)]
        parser = gateway.jvm.org.apache.ignite.ml.xgboost.parser.XGModelParser()
        
        super(XGBoostDistributedModel, self).__init__(ignite, reader, parser, instances, max_per_node)

    def predict(self, X):
        keys = gateway.jvm.java.util.HashMap()
        data = []

        idx = 0
        for key in X:
            keys[key] = idx
            idx = idx + 1
            data.append(X[key])

        java_array = Utils.to_java_double_array(data)
        java_vector_utils = gateway.jvm.org.apache.ignite.ml.math.primitives.vector.VectorUtils

        X = gateway.jvm.org.apache.ignite.ml.math.primitives.vector.impl.DelegatingNamedVector(java_vector_utils.of(java_array), keys)

        res = self.proxy[0].predict(X)
        # This if handles 'future' response.
        if hasattr(res, 'get') and callable(res.get):
            res = res.get()
        return res

class IgniteDistributedModel(DistributedModel):
    """Ignite distributed model.

    Parameters
    ----------
    ignite : Ignite instance.
    mdl : Model.
    instances : Number of instances.
    max_per_node : Max number of instance per node.
    """
    def __init__(self, ignite, mdl, instances=1, max_per_node=1):
        """Constructs a new instance of Ignite distributed model.

        Parameters
        ----------
        ignite : Ignite instance.
        reader : Model reader.
        parser : Model parser.
        instances : Number of worker instances.
        max_per_node : Max number of worker instances per ignite node.
        """
        if isinstance(mdl.proxy, list):
            reader = [gateway.jvm.org.apache.ignite.ml.inference.reader.InMemoryModelReader(p) for p in mdl.proxy]
        else:
            reader = [gateway.jvm.org.apache.ignite.ml.inference.reader.InMemoryModelReader(mdl.proxy)]

        parser = gateway.jvm.org.apache.ignite.ml.inference.parser.IgniteModelParser()

        super(IgniteDistributedModel, self).__init__(ignite, reader, parser, instances, max_per_node)



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Regression trainers.
"""

import numpy as np

from ..common import Model
from ..common import SupervisedTrainer
from ..common import Proxy
from ..common import Utils
from ..common import LearningEnvironmentBuilder

from ..core import Cache

from ..common import gateway

class RegressionTrainer(SupervisedTrainer, Proxy):
    """Regression.
    """
    def __init__(self, proxy, multiple_labels=False, accepts_matrix=False):
        """Constructs a new instance of regression trainer.
        """
        self.multiple_labels = multiple_labels
        self.accepts_matrix = accepts_matrix
        Proxy.__init__(self, proxy)

    def fit(self, X, y=None):
        if isinstance(X, Cache):
            if y is not None:
                raise Exception("Second argument (y) is unexpected in case the first parameters is cache.")
            return self.fit_on_cache(X)

        X = np.array(X)
        y = np.array(y)

        # Check dimensions: we expected to have 2-dim X and y arrays.
        if X.ndim == 1:
            X = X.reshape(X.shape[0], 1)
        elif X.ndim > 2:
            raise Exception("X has unexpected dimension [dim=%d]" % X.ndim)

        if y.ndim == 1:
            y = y.reshape(y.shape[0], 1)
        elif y.ndim > 2:
            raise Exception("y has unexpected dimension [dim=%d]" % y.ndim)

        X_java = Utils.to_java_double_array(X)

        # We have two types of models: first type can accept multiple labels, second can't.
        if self.multiple_labels:        
            y_java = Utils.to_java_double_array(y)
            java_model = self.proxy.fit(X_java, y_java, None)
            return Model(java_model, self.accepts_matrix)
        else:
            java_models = []
            # Here we need to prepare a model for each y column.
            for i in range(y.shape[1]):
                y_java = Utils.to_java_double_array(y[:,i])
                java_model = self.proxy.fit(X_java, y_java, None)
                java_models.append(java_model)
            return Model(java_models, self.accepts_matrix)

    def fit_on_cache(self, cache):
        if not isinstance(cache, Cache):
            raise Exception("Unexpected type of cache (%s)." % type(cache))

        java_model = self.proxy.fitOnCache(cache.proxy, cache.cache_filter, Proxy.proxy_or_none(cache.preprocessor))
        return Model(java_model, self.accepts_matrix)


class DecisionTreeRegressionTrainer(RegressionTrainer):
    """DecisionTree regression trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(),
                 max_deep=5,
                 min_impurity_decrease=0.0, compressor=None, use_index=True):
        """Constructs a new instance of DecisionTree regression trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        max_deep : Max deep.
        min_impurity_decrease : Min impurity decrease.
        compressor : Compressor.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.tree.DecisionTreeRegressionTrainer(max_deep, min_impurity_decrease, compressor)
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))
        proxy.withUsingIdx(use_index)

        RegressionTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class KNNRegressionTrainer(RegressionTrainer):
    """KNN regression trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder()):
        """Constructs a new instance of linear regression trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.knn.regression.KNNRegressionTrainer()
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))

        RegressionTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class LinearRegressionTrainer(RegressionTrainer):
    """Linear regression trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder()):
        """Constructs a new instance of linear regression trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.regressions.linear.LinearRegressionLSQRTrainer()
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))

        RegressionTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class RandomForestRegressionTrainer(RegressionTrainer):
    """RandomForest classification trainer.
    """
    def __init__(self, features, env_builder=LearningEnvironmentBuilder(),
                 trees=1, sub_sample_size=1.0, max_depth=5,
                 min_impurity_delta=0.0, seed=None):
        """Constructs a new instance of RandomForest classification trainer.

        Parameters
        ----------
        features : Number of features.
        env_builder : Environment builder.
        trees : Number of trees.
        sub_sample_size : Sub sample size.
        max_depth : Max depth.
        min_impurity_delta : Min impurity delta.
        seed : Seed.
        """
        metas = gateway.jvm.java.util.ArrayList()
        for i in range(features):
            meta = gateway.jvm.org.apache.ignite.ml.dataset.feature.FeatureMeta(None, i, False)
            metas.add(meta)

        proxy = gateway.jvm.org.apache.ignite.ml.tree.randomforest.RandomForestClassifierTrainer(metas)
        proxy.withEnvironmentBuilder(env_builder.proxy)
        proxy.withAmountOfTrees(trees)
        proxy.withSubSampleSize(sub_sample_size)
        proxy.withMaxDepth(max_depth)
        proxy.withMinImpurityDelta(min_impurity_delta)
        if seed is not None:
            proxy.withSeed(seed)

        RegressionTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class MLPArchitecture(Proxy):
    """MLP architecture.
    """
    def __init__(self, input_size):
        """Constructs a new instance of MLP architecture.

        Parameters
        ----------
        input_size : Input size.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.nn.architecture.MLPArchitecture(input_size)
        Proxy.__init__(self, proxy)

    def with_layer(self, neurons, has_bias=True, activator='sigmoid'):
        """Add layer.

        Parameters
        ----------
        neurons : Number of neurons.
        has_bias : Has bias or not (default value is True).
        activator : Activation function ('sigmoid', 'relu' or 'linear', default value is 'sigmoid')
        """
        java_activator = None
        if activator == 'sigmoid':
            java_activator = gateway.jvm.org.apache.ignite.ml.nn.Activators.SIGMOID
        elif activator == 'relu':
            java_activator = gateway.jvm.org.apache.ignite.ml.nn.Activators.RELU
        elif activator == 'linear':
            java_activator = gateway.jvm.org.apache.ignite.ml.nn.Activators.LINEAR
        else:
            raise Exception("Unknown activator: %s" % activator)

        self.proxy = self.proxy.withAddedLayer(neurons, has_bias, java_activator)

        return self

class MLPRegressionTrainer(RegressionTrainer):
    """MLP regression trainer.
    """
    def __init__(self, arch, env_builder=LearningEnvironmentBuilder(), loss='mse',
                 learning_rate=0.1, max_iter=1000, batch_size=100, loc_iter=10, seed=None):
        """Constructs a new instance of MLP regression trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        arch : Architecture.
        loss : Loss function ('mse', 'log', 'l2', 'l1' or 'hinge', default value is 'mse').
        update_strategy : Update strategy.
        max_iter : Max number of iterations.
        batch_size : Batch size.
        loc_iter : Number of local iterations.
        seed : Seed.
        """
        java_loss = None
        if loss == 'mse':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.MSE
        elif loss == 'log':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.LOG
        elif loss == 'l2':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.L2
        elif loss == 'l1':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.L1
        elif loss == 'hinge':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.HINGE
        else:
            raise Exception('Unknown loss: %s' % loss)

        proxy = gateway.jvm.org.gridgain.ml.python.PythonMLPDatasetTrainer(arch.proxy, java_loss, learning_rate, max_iter, batch_size, loc_iter, seed)
        RegressionTrainer.__init__(self, proxy, True, True)



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Preprocessors.
"""

import numpy as np

from ..common import UnsupervisedTrainer
from ..common import Proxy
from ..common import Utils
from ..core import Cache
from ..common import gateway

class PreprocessingModel(Proxy):
    """Preprocessing model.
    """
    def __init__(self, proxy):
        """Constructs a new instance of preprocessing model.
        """
        Proxy.__init__(self, proxy)

    def transform(self, X):
        X = np.array(X)

        if X.ndim == 1:
            X = X.reshape(X.shape[0], 1)
        elif X.ndim > 2:
            raise Exception("X has unexpected dimension [dim=%d]" % X.ndim)

        transformations = np.array([self.__transform(x) for x in X])
        if transformations.ndim == 2 and transformations.shape[1] == 1:
            transformations = np.hstack(transformations)

        return transformations

    def __transform(self, X):
        java_array = Utils.to_java_double_array(X)
        res = self.proxy.apply(0, java_array)
        return [res.get(i) for i in range(res.size())]


class PreprocessingTrainer(UnsupervisedTrainer):
    """Preprocessing trainer.
    """
    def __init__(self, proxy):
        """Constructs a new instance of PreprocessingTrainer.

        Parameters
        ----------
        proxy : Java proxy.
        """
        Proxy.__init__(self, proxy)

    def fit(self, X):
        if isinstance(X, Cache):
            return self.fit_on_cache(X)

        X_java = Utils.to_java_double_array(X)
        java_model = self.proxy.fit(X_java, None)

        return PreprocessingModel(java_model)

    def fit_on_cache(self, cache):
        java_model = self.proxy.fitOnCache(cache.proxy, Proxy.proxy_or_none(cache.preprocessor))

        return PreprocessingModel(java_model)

class MinMaxScalerTrainer(PreprocessingTrainer):
    """Min-max scaler trainer.
    """
    def __init__(self):
        """Constructs a new instance of min-max scaler trainer.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.minmaxscaling.MinMaxScalerTrainer()

        PreprocessingTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonPreprocessingTrainer(proxy))

class MaxAbsScalerTrainer(PreprocessingTrainer):
    """Max absolute scaler trainer.
    """
    def __init__(self):
        """Constructs a new instance of max absolute scaler trainer.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.maxabsscaling.MaxAbsScalerTrainer()

        PreprocessingTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonPreprocessingTrainer(proxy))

class BinarizationTrainer(PreprocessingTrainer):
    """Binarization trainer.
    """
    def __init__(self, threshold=0.0):
        """Constructs a new instance of binarization trainer.

        Parameters
        ----------
        threshold : Threshold (Default value is 0).
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.binarization.BinarizationTrainer()
        proxy.withThreshold(threshold)

        PreprocessingTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonPreprocessingTrainer(proxy))

class EncoderTrainer(PreprocessingTrainer):
    """Encoder trainer.
    """
    def __init__(self, encoded_features=[], encoder_indexing_strategy='frequency_desc', encoder_type='one_hot'):
        """Constructs a new instance of encoder trainer.

        Parameters
        ----------
        encoder_features : Encoded features (Default value is []).
        encoder_indexing_strategy : Encoder indexing strategy ('frequency_desc', 'frequency_asc', default value is 'frequency_desc').
        encoder_type : Encoder type ('one_hot', 'string', default value is 'one_hot').
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.encoding.EncoderTrainer()
        if not encoded_features:
            encoded_features = []
        for encoded_feature in encoded_features:
            proxy.withEncodedFeature(encoded_feature)

        java_encoder_indexing_strategy = None
        if encoder_indexing_strategy == 'frequency_desc':
            java_encoder_indexing_strategy = gateway.jvm.org.apache.ignite.ml.preprocessing.encoding.EncoderSortingStrategy.FREQUENCY_DESC
        elif encoder_indexing_strategy == 'frequency_asc':
            java_encoder_indexing_strategy = gateway.jvm.org.apache.ignite.ml.preprocessing.encoding.EncoderSortingStrategy.FREQUENCY_ASC
        elif encoder_indexing_strategy:
            raise Exception('Unknown encoder indexing strategy: %s' % encoder_indexing_strategy)
        proxy.withEncoderIndexingStrategy(java_encoder_indexing_strategy)
        
        java_encoder_type = None
        if encoder_type == 'one_hot':
            java_encoder_type = gateway.jvm.org.apache.ignite.ml.preprocessing.encoding.EncoderType.ONE_HOT_ENCODER
        elif encoder_type == 'string':
            java_encoder_type = gateway.jvm.org.apache.ignite.ml.preprocessing.encoding.EncoderType.STRING_ENCODER
        elif encoder_type:
            raise Exception("Unknown encoder type: %s" % encoder_type)
        proxy.withEncoderType(java_encoder_type)

        self.proxy = gateway.jvm.org.gridgain.ml.python.PythonEncoderPreprocessingTrainer(proxy)

    def fit(self, X):
        if isinstance(X, Cache):
            return self.fit_on_cache(X)

        X_java = Utils.to_java_double_array(X)
        java_model = self.proxy.fit(X_java)

        return PreprocessingModel(java_model)

    def fit_on_cache(self, cache):
        if cache.preprocessor:
            raise Exception("Encoder preprocessing trainer doesn't support upstream preprocessors")

        java_model = self.proxy.fitOnCache(cache.proxy)

        return PreprocessingModel(java_model)

class ImputerTrainer(PreprocessingTrainer):
    """Imputer trainer.
    """
    def __init__(self, imputing_strategy='mean'):
        """Constructs a new instance of imputer trainer.

        Parameters
        ----------
        imputing_strategy : Imputing strategy ('mean', 'most_frequent', default value is 'mean').
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.imputing.ImputerTrainer()

        java_imputing_strategy = None
        if imputing_strategy == 'mean':
            java_imputing_strategy = gateway.jvm.org.apache.ignite.ml.preprocessing.imputing.ImputingStrategy.MEAN
        elif imputing_strategy == 'most_frequent':
            java_imputing_strategy = gateway.jvm.org.apache.ignite.ml.preprocessing.imputing.ImputingStrategy.MOST_FREQUENT
        elif imputing_strategy:
            raise Exception("Unknown imputing strategy: %s" % imputing_strategy)
        proxy.withImputingStrategy(java_imputing_strategy)

        PreprocessingTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonPreprocessingTrainer(proxy))

class NormalizationTrainer(PreprocessingTrainer):
    """Normalization trainer.
    """
    def __init__(self, p=2):
        """Constructs a new instance of normalization trainer.

        Parameters
        ----------
        p : Degree of L space parameter value.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.normalization.NormalizationTrainer()
        proxy.withP(p)

        PreprocessingTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonPreprocessingTrainer(proxy))

class StandardScalerTrainer(PreprocessingTrainer):
    """Standard scaler trainer.
    """
    def __init__(self):
        """Constructs a new instance of standard scaler trainer.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.preprocessing.standardscaling.StandardScalerTrainer()

        PreprocessingTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonPreprocessingTrainer(proxy))



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
from numbers import Number

from ..common import SupervisedTrainer
from ..common import Proxy
from ..common import Utils
from ..common import LearningEnvironmentBuilder

from ..common import gateway

from ..core import Cache

__evaluator = gateway.jvm.org.gridgain.ml.python.PythonEvaluator

def __evaluate_regression(cache, mdl, preprocessor=None):
    if not isinstance(cache, Cache):
        raise Exception("Unexpected type of cache (%s)." % type(cache))

    return __evaluator.evaluateRegression(cache.proxy, None, mdl.proxy, preprocessor)

def __evaluate_classification(cache, mdl, preprocessor=None):
    if not isinstance(cache, Cache):
        raise Exception("Unexpected type of cache (%s)." % type(cache))

    return __evaluator.evaluateClassification(cache.proxy, None, mdl.proxy, preprocessor)

def accuracy_score(cache, mdl, preprocessor=None):
    """Calculate accuracy score (classification metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    classification_metrics = __evaluate_classification(cache, mdl, preprocessor)
    return classification_metrics.accuracy()

def balanced_accuracy_score(cache, mdl, preprocessor=None):
    """Calculate balanced accuracy score (classification metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    classification_metrics = __evaluate_classification(cache, mdl, preprocessor)
    return classification_metrics.balancedAccuracy()

def precision_score(cache, mdl, preprocessor=None):
    """Calculate precision score (classification metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    classification_metrics = __evaluate_classification(cache, mdl, preprocessor)
    return classification_metrics.precision()

def recall_score(cache, mdl, preprocessor=None):
    """Calculate recall score (classification metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    classification_metrics = __evaluate_classification(cache, mdl, preprocessor)
    return classification_metrics.recall()

def f1_score(cache, mdl, preprocessor=None):
    """Calculate f1 score (classification metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    classification_metrics = __evaluate_classification(cache, mdl, preprocessor)
    return classification_metrics.f1Score()

def mae_score(cache, mdl, preprocessor=None):
    """Calculate mean absolute error score (regression metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    regression_metrics = __evaluate_regression(cache, mdl, preprocessor)
    return regression_metrics.mae()

def mse_score(cache, mdl, preprocessor=None):
    """Calculate mean squared error score (regression metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    regression_metrics = __evaluate_regression(cache, mdl, preprocessor)
    return regression_metrics.mse()

def rss_score(cache, mdl, preprocessor=None):
    """Calculate residual sum of squares score (regression metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    regression_metrics = __evaluate_regression(cache, mdl, preprocessor)
    return regression_metrics.rss()

def rmse_score(cache, mdl, preprocessor=None):
    """Calculate root mean squared error score (regression metric).

    Parameters
    ----------
    cache : Cache or cache view (cache with filter).
    mdl : Model.
    preprocessor : Preprocessor.
    """
    regression_metrics = __evaluate_regression(cache, mdl, preprocessor)
    return regression_metrics.rmse()



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Common classes.
"""

from abc import abstractmethod

import os
import sys
import numpy as np
from py4j.java_gateway import JavaGateway

if 'IGNITE_HOME' in os.environ:
    ignite_home = os.environ['IGNITE_HOME']

    libs_jar = []
    for f in os.listdir(ignite_home + '/libs'):
        if f.endswith('.jar'):
            libs_jar.append(ignite_home + '/libs/' + f)    
        if os.path.isdir(ignite_home + '/libs/' + f):
            for fi in os.listdir(ignite_home + '/libs/' + f):
                if fi.endswith('.jar'):
                    libs_jar.append(ignite_home + '/libs/' + f + '/' + fi)

    optional_libs_jar = []
    for opt in os.listdir(ignite_home + '/libs/optional'):
        for f in os.listdir(ignite_home + '/libs/optional/' + opt):
            if f.endswith('.jar'):
                optional_libs_jar.append(ignite_home + '/libs/optional/' + opt + '/' + f)

    classpath = ':'.join(libs_jar + optional_libs_jar)

    gateway = JavaGateway.launch_gateway(classpath=classpath, die_on_exit=True)
else:
    # To build documentation.
    gateway = None

class Utils:
    """Util class.
    """
    def to_java_double_array(array):
        """Converts python array into java double array.
        """
        array = np.array(array)
        java_array = gateway.new_array(gateway.jvm.double, *array.shape)
        Utils.__to_java_double_array_backtrack(array, java_array)
        return java_array

    def from_java_double_array(java_array):
        """Converts java double array into python array.
        """
        array = np.zeros(len(java_array))
        for i in range(len(java_array)):
            array[i] = java_array[i]
        return array

    def __to_java_double_array_backtrack(array, java_array):
        if array.ndim == 0:
            raise Exception("Array is scalar [dim=%d]" % array.ndim)

        for i in range(array.shape[0]):
            if array.ndim == 1:
                if array[i] is not None:
                    java_array[i] = float(array[i])
                else:
                    java_array[i] = float('NaN')
            else:
                Utils.__to_java_double_array_backtrack(array[i], java_array[i])

class Proxy:
    """Proxy class for Java object.
    """
    def __init__(self, proxy):
        """Constructs a new instance of proxy class for Java object.
        """
        self.proxy = proxy

    def proxy_or_none(proxy):
        """Returns proxy of the given object or None of object is None itself.
        """
        if proxy:
            return proxy.proxy
        else:
            return None

class LearningEnvironmentBuilder(Proxy):
    """Learning environment builder.
    """
    def __init__(self):
        """Constructs a new instance of learning environemtn builder.
        """
        if gateway:
            java_proxy = gateway.jvm.org.apache.ignite.ml.environment.LearningEnvironmentBuilder.defaultBuilder()
            Proxy.__init__(self, java_proxy)

class Model(Proxy):
    """Model.
    """
    def __init__(self, proxy, accepts_matrix):
        """Constructs a new instance of regression model.

        Parameters
        ----------
        proxy : Proxy object that represents Java model.
        accept_matrix : Flag that identifies if model accepts matrix or vector.
        """
        self.accepts_matrix = accepts_matrix
        Proxy.__init__(self, proxy)

    def predict(self, X):
        """Predicts a result.

        Parameters
        ----------

        X : Features.
        """
        X = np.array(X)

        if X.ndim == 1:
            X = X.reshape(X.shape[0], 1)
        elif X.ndim > 2:
            raise Exception("X has unexpected dimension [dim=%d]" % X.ndim)

        # Check if model accepts multiple objects for inference.
        if self.accepts_matrix:
            java_array = Utils.to_java_double_array(X)
            java_matrix = gateway.jvm.org.apache.ignite.ml.math.primitives.matrix.impl.DenseMatrix(java_array)
            # Check if model is a single model or model-per-label.
            if isinstance(self.proxy, list):
                predictions = np.array([mdl.predict(java_matrix) for mdl in self.proxy])
            else:
                res = self.proxy.predict(java_matrix)
                rows = res.rowSize()
                cols = res.columnSize()
                predictions = np.zeros((rows, cols))
                for i in range(rows):
                    for j in range(cols):
                        predictions[i, j] = res.get(i, j)
        else:
            predictions = []
            for i in range(X.shape[0]):
                java_array = Utils.to_java_double_array(X[i])
                java_vector_utils = gateway.jvm.org.apache.ignite.ml.math.primitives.vector.VectorUtils
                # Check if model is a single model or model-per-label.
                if isinstance(self.proxy, list):
                    def parse_response(m):
                        res = m.predict(java_vector_utils.of(java_array))
                        # This if handles 'future' response.
                        if hasattr(res, 'get') and callable(res.get):
                            res = res.get()
                        return res
                    prediction = [parse_response(mdl) for mdl in self.proxy]
                else:
                    prediction = [self.proxy.predict(java_vector_utils.of(java_array))]
                predictions.append(prediction)
            predictions = np.array(predictions)

        if predictions.ndim == 2 and predictions.shape[1] == 1:
            predictions = np.hstack(predictions)

        return predictions

class SupervisedTrainer:
    """Supervised trainer.
    """
    @abstractmethod
    def fit(self, X, y=None):
        """Trains model based on data.

        Parameters
        ----------
        X : x.
        y : y.
        """
        raise Exception("Not implemented")

    @abstractmethod
    def fit_on_cache(self, cache):
        """Trains model based on data.

        Parameters
        ----------
        cache : Apache Ignite cache.
        """
        raise Exception("Not implemented")

class UnsupervisedTrainer:
    """Unsupervised trainer.
    """
    @abstractmethod
    def fit(self, X):
        """Trains model based on data.

        Parameters
        ----------
        X : x.
        """
        raise Exception("Not implemented")

    @abstractmethod
    def fit_on_cache(self, cache):
        """Trains model based on data.

        Parameters
        ----------
        cache : Apache Ignite cache.
        """
        raise Exception("Not implemented")



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Ignite inference functionality.
"""

from ..common import Model
from ..common import Proxy
from ..common import Utils
from ..common import gateway
from copy import copy
from numbers import Number

def save_model(mdl, dst, ignite=None):
    if not isinstance(mdl.proxy, list):
        gateway.jvm.org.gridgain.ml.python.PythonModelSaver.save(mdl.proxy, dst, ignite.ignite)

    if len(mdl.proxy) != 1:
        raise Exception("Models with multiple inner models are not supported")

    if mdl.accepts_matrix:
        raise Exception("Models that accept matrix are not supported")

    gateway.jvm.org.gridgain.ml.python.PythonModelSaver.save(mdl.proxy[0], dst, ignite.ignite)

def read_model(src, ignite=None):
    proxy = [gateway.jvm.org.gridgain.ml.python.PythonModelSaver.read(src, ignite.ignite)]

    return Model(proxy, False)



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Ignite cache API.
"""

from ..common import Proxy
from ..common import Utils

from ..common import gateway

import pandas as pd
import numpy as np

class Ignite:
 
    def __init__(self, cfg=None):
        """
        Constructs a new instance of Ignite that is required to work with Cache, IGFS storage and
        distributed inference.

        :param cfg: Path to Apache Ignite configuration file.
        """
        self.cfg = cfg
   
    def get_cache(self, name):
        """
        Returns existing Apache Ignite Cache by name. This module is built with assumption that Ignite
        Cache contains integer keys and double[] values.

        :param name: Name of the Apache Ignite cache.
        """
        if self.ignite is None:
            raise Exception("Use Ignite() inside with.. as.. command.")
        java_cache = self.ignite.cache(name)
        return Cache(java_cache)

    def create_cache(self, name, excl_neighbors=False, parts=10):
        """
        Creates a new Apache Ignite Cache using specified name and configuration. This module is built with
        assumption that Ignite Cache contains integer keys and double[] values.

        :param name: Name of the Apache Ignite cache,
        :param excl_neighbors: (optional, False by default) exclude neighbours,
        :param parts: (optional, 10 by default) number of partitions.
        """
        if self.ignite is None:
            raise Exception("Use Ignite() inside with.. as.. command.")
        affinity = gateway.jvm.org.apache.ignite.cache.affinity.rendezvous.RendezvousAffinityFunction(excl_neighbors, parts)
        cc = gateway.jvm.org.apache.ignite.configuration.CacheConfiguration()
        cc.setName(name)
        cc.setAffinity(affinity)
        java_cache = self.ignite.createCache(cc)
        return Cache(java_cache)

    def __enter__(self):
        if self.cfg is not None:
            self.ignite = gateway.jvm.org.apache.ignite.Ignition.start(self.cfg)
        else:
            self.ignite = gateway.jvm.org.apache.ignite.Ignition.start()
        return self

    def __exit__(self, t, v, trace):
        if self.ignite is not None:
            self.ignite.close()

class Cache(Proxy):
    """Internal constructor that creates a wrapper of Apache Ignite cache. User is expected to use Ignite
    object to create cache instead of this constructor.
    """
    def __init__(self, proxy, cache_filter=None, preprocessor=None):
        """
        Constructs a wrapper of Apache Ignite cache. It's internal method, user is expected to use Ignite
        methods to create or get cache.

        :param proxy: Py4J proxy that represents Apache Ignite Cache,
        :param cache_filter: Py4J proxy that represents filter,
        :param preprocessor: Py4J proxy that represents preprocessor.
        """
        Proxy.__init__(self, proxy)

        self.cache_filter = cache_filter
        self.preprocessor = preprocessor

    def __delitem__(self, key):
        raise Exception("Not implemented!")

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.get(key)
        elif isinstance(key, slice):
            result = []
            length = len(self)
            for k in range(*key.indices(length)):
                result.append(self.get(k))
            return np.array(result)
        else:
            raise Exception("Unexpected type of key (%s)." % type(key))

    def __setitem__(self, key, value):
        raise Exception("Not implemented!")

    def get(self, key):
        """
        Returns value (float array) by the specified key.

        :param key: Key to be taken from cache.
        """
        java_array = self.proxy.get(key)
        return Utils.from_java_double_array(java_array)

    def put(self, key, value):
        """
        Puts value (float array) by the specified key.

        :param key: Key to be put into cache,
        :param value: value to be taken from cache.
        """
        value = Utils.to_java_double_array(value)
        self.proxy.put(key, value)

    def head(self, n=5):
        """
        Returns top N elements represented as a pandas dataset.

        :param n: Number of rows to be returned.
        """
        scan_query = gateway.jvm.org.apache.ignite.cache.query.ScanQuery()
        
        if self.cache_filter is not None:
            scan_query.setFilter(self.cache_filter)

        cursor = self.proxy.query(scan_query)
        iterator = cursor.iterator()
        
        data = []
        while iterator.hasNext() and n != 0:
            entry = iterator.next()
            key = entry.getKey()
            value = entry.getValue()

            if self.preprocessor is not None:
                initial_array = Utils.from_java_double_array(value)
                preprocessed_value = self.preprocessor.proxy.apply(key, value)
                value = preprocessed_value.features().asArray()
                array = Utils.from_java_double_array(value)
                array = np.hstack((array, initial_array[-1]))
            else:
                array = Utils.from_java_double_array(value)
    
            data.append(array)
            n = n - 1

        return pd.DataFrame(data)

    def transform(self, preprocessor):
        """
        Transform this cache using specfied preprocessor.

        :param preprocessor: Preprocessor to be used to transform cache.
        """
        return Cache(self.proxy, self.cache_filter, preprocessor)

    def filter(self, cache_filter):
        """
        Filters this cache using specified filter.

        :param filter: Filter to be used to filter cache.
        """
        return Cache(self.proxy, cache_filter, self.preprocessor)

    def __len__(self):
        return self.proxy.size(gateway.new_array(gateway.jvm.org.apache.ignite.cache.CachePeekMode, 0))



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Classification trainers.
"""

import numpy as np

from ..common import Model
from ..common import SupervisedTrainer
from ..common import Proxy
from ..common import Utils
from ..common import LearningEnvironmentBuilder

from ..common import gateway
from ..core import Cache

class ClassificationTrainer(SupervisedTrainer, Proxy):
    """Classification trainer.
    """
    def __init__(self, proxy):
        """Constructs a new instance of classification trainer.
        """
        Proxy.__init__(self, proxy)

    def fit(self, X, y=None):
        if isinstance(X, Cache):
            if y is not None:
                raise Exception("Second argument (y) is unexpected in case the first parameters is cache.")
            return self.fit_on_cache(X)

        X_java = Utils.to_java_double_array(X)
        y_java = Utils.to_java_double_array(y)

        java_model = self.proxy.fit(X_java, y_java, None)

        return Model(java_model, False)

    def fit_on_cache(self, cache):
        if not isinstance(cache, Cache):
            raise Exception("Unexpected type of cache (%s)." % type(cache))

        java_model = self.proxy.fitOnCache(cache.proxy, cache.cache_filter, Proxy.proxy_or_none(cache.preprocessor))
        return Model(java_model, False)

class ANNClassificationTrainer(ClassificationTrainer):
    """ANN classification trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(), k=2,
                 max_iter=10, eps=1e-4, distance='euclidean'):
        """Constructs a new instance of ANN classification trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        k : Number of clusters.
        max_iter : Max number of iterations.
        eps : Epsilon, delta of convergence.
        distance : Distance measure ('euclidean', 'hamming', 'manhattan').
        """
        proxy = gateway.jvm.org.apache.ignite.ml.knn.ann.ANNClassificationTrainer()

        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))
        proxy.withK(k)
        proxy.withMaxIterations(max_iter)
        proxy.withEpsilon(eps)

        java_distance = None
        if distance == 'euclidean':
            java_distance = gateway.jvm.org.apache.ignite.ml.math.distances.EuclideanDistance()
        elif distance == 'hamming':
            java_distance = gateway.jvm.org.apache.ignite.ml.math.distances.HammingDistance()
        elif distance == 'manhattan':
            java_distance = gateway.jvm.org.apache.ignite.ml.math.distances.ManhattanDistance()
        elif distance:
            raise Exception("Unknown distance type: %s" % distance)
        proxy.withDistance(java_distance)

        ClassificationTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class DecisionTreeClassificationTrainer(ClassificationTrainer):
    """DecisionTree classification trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(), max_deep=5,
                 min_impurity_decrease=0.0, compressor=None, use_index=True):
        """Constructs a new instance of DecisionTree classification trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        max_deep : Max deep.
        min_impurity_decrease : Min impurity decrease.
        compressor : Compressor.
        use_index : Use index.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.tree.DecisionTreeClassificationTrainer(max_deep, min_impurity_decrease, compressor)

        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))
        proxy.withUseIndex(use_index)

        ClassificationTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class KNNClassificationTrainer(ClassificationTrainer):
    """KNN classification trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder()):
        """Constructs a new instance of KNN classification trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.knn.classification.KNNClassificationTrainer()
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))

        ClassificationTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class LogRegClassificationTrainer(ClassificationTrainer):
    """LogisticRegression classification trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(), max_iter=100,
                 batch_size=100, max_loc_iter=100, seed=1234):
        """Constructs a new instance of LogisticRegression classification trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        max_iter : Max number of iterations.
        batch_size : Batch size.
        max_loc_iter : Max number of local iterations.
        update_strategy : Update strategy.
        seed : Seed.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.regressions.logistic.LogisticRegressionSGDTrainer()

        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))

        proxy.withMaxIterations(max_iter)
        proxy.withBatchSize(batch_size)
        proxy.withLocIterations(max_loc_iter)
        if seed is not None:
            proxy.withSeed(seed)

        ClassificationTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class RandomForestClassificationTrainer(ClassificationTrainer):
    """RandomForest classification trainer.
    """
    def __init__(self, features, env_builder=LearningEnvironmentBuilder(),
                 trees=1, sub_sample_size=1.0, max_depth=5,
                 min_impurity_delta=0.0, seed=None):
        """Constructs a new instance of RandomForest classification trainer.

        Parameters
        ----------
        features : Number of features.
        env_builder : Environment builder.
        trees : Number of trees.
        sub_sample_size : Sub sample size.
        max_depth : Max depth.
        min_impurity_delta : Min impurity delta.
        seed : Seed.
        """
        metas = gateway.jvm.java.util.ArrayList()
        for i in range(features):
            meta = gateway.jvm.org.apache.ignite.ml.dataset.feature.FeatureMeta(None, i, False)
            metas.add(meta)

        proxy = gateway.jvm.org.apache.ignite.ml.tree.randomforest.RandomForestClassifierTrainer(metas)
        proxy.withEnvironmentBuilder(env_builder.proxy)
        proxy.withAmountOfTrees(trees)
        proxy.withSubSampleSize(sub_sample_size)
        proxy.withMaxDepth(max_depth)
        proxy.withMinImpurityDelta(min_impurity_delta)
        if seed is not None:
            proxy.withSeed(seed)

        ClassificationTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class SVMClassificationTrainer(ClassificationTrainer):
    """SVM classification trainer.
    """
    def __init__(self, env_builder=LearningEnvironmentBuilder(), l=0.4, max_iter=200, max_local_iter=100, seed=1234):
        """Constructs a new instance of SVM classification trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        l : Lambda.
        max_iter : Max number of iterations.
        max_loc_iter : Max number of local iterations.
        seed : Seed.
        """
        proxy = gateway.jvm.org.apache.ignite.ml.svm.SVMLinearClassificationTrainer()    
        proxy.withEnvironmentBuilder(Proxy.proxy_or_none(env_builder))
        proxy.withLambda(l)
        proxy.withAmountOfIterations(max_iter)
        proxy.withAmountOfLocIterations(max_local_iter)
        if seed is not None:
            proxy.withSeed(seed)

        ClassificationTrainer.__init__(self, gateway.jvm.org.gridgain.ml.python.PythonDatasetTrainer(proxy))

class MLPClassificationTrainer(ClassificationTrainer):
    """MLP regression trainer.
    """
    def __init__(self, arch, env_builder=LearningEnvironmentBuilder(), loss='mse',
                 learning_rate=0.1, max_iter=1000, batch_size=100, loc_iter=10, seed=None):
        """Constructs a new instance of MLP regression trainer.

        Parameters
        ----------
        env_builder : Environment builder.
        arch : Architecture.
        loss : Loss function ('mse', 'log', 'l2', 'l1' or 'hinge', default value is 'mse').
        update_strategy : Update strategy.
        max_iter : Max number of iterations.
        batch_size : Batch size.
        loc_iter : Number of local iterations.
        seed : Seed.
        """
        java_loss = None
        if loss == 'mse':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.MSE
        elif loss == 'log':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.LOG
        elif loss == 'l2':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.L2
        elif loss == 'l1':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.L1
        elif loss == 'hinge':
            java_loss = gateway.jvm.org.apache.ignite.ml.optimization.LossFunctions.HINGE
        else:
            raise Exception('Unknown loss: %s' % loss)

        proxy = gateway.jvm.org.gridgain.ml.python.PythonMLPDatasetTrainer(arch.proxy, java_loss, learning_rate, max_iter, batch_size, loc_iter, seed)
        RegressionTrainer.__init__(self, proxy, True)



#
# Copyright 2019 GridGain Systems, Inc. and Contributors.
#
# Licensed under the GridGain Community Edition License (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gridgain.com/products/software/community-edition/gridgain-community-edition-license
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..core import Cache
from ..common import gateway
from ..common import Utils

def train_test_split(cache, test_size=0.25, train_size=0.75, random_state=None):
    """Splits given cache on two parts: test and train with given sizes.

    Parameters
    ----------
    cache : Ignite cache.
    test_size : Test size.
    train_size : Train size.
    random_state : Random state.
    """
    if not isinstance(cache, Cache):
        raise Exception("Unexpected type of cache (%s)." % type(cache))    

    split = gateway.jvm.org.apache.ignite.ml.selection.split.TrainTestDatasetSplitter().split(train_size, test_size)
    train_filter = split.getTrainFilter()
    test_filter = split.getTestFilter()
    return (cache.filter(train_filter), cache.filter(test_filter))

def cross_val_score(trainer, cache, cv=5, scoring='accuracy'):
    """Makes cross validation for given trainer, cache and scoring.

    Parameters
    ----------
    trainer : Trainer.
    cache : Cache.
    cv : Number of folds.
    scoring : Metric to be scored.
    """
    if not isinstance(cache, Cache):
        raise Exception("Unexpected type of cache (%s)." % type(cache))

    if scoring == 'accuracy':
        metric = gateway.jvm.org.apache.ignite.ml.selection.scoring.metric.classification.Accuracy()
    else:
        raise Exception("Unsupported type of scoring metric: %s" % scoring)

    res = gateway.jvm.org.gridgain.ml.python.PythonCrossValidation.score(trainer.proxy.getDelegate(), metric, cache.proxy, cv)

    return Utils.from_java_double_array(res)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import setuptools
import sys

PYTHON_REQUIRED = (3, 4)
PYTHON_INSTALLED = sys.version_info[:2]

if PYTHON_INSTALLED < PYTHON_REQUIRED:
    sys.stderr.write('''

`ggml` is not compatible with Python {}.{}!
Use Python {}.{} or above.

'''.format(
        PYTHON_INSTALLED[0],
        PYTHON_INSTALLED[1],
        PYTHON_REQUIRED[0],
        PYTHON_REQUIRED[1],
    )
    )
    sys.exit(1)

def is_a_requirement(line):
    return not any([
        line.startswith('#'),
        line.startswith('-r'),
        len(line) == 0,
        ])


requirement_sections = [
    'install',
    'setup',
    'tests',
    'docs',
]
requirements = defaultdict(list)

for section in requirement_sections:
    with open(
            'requirements/{}.txt'.format(section),
            'r',
            encoding='utf-8',
    ) as requirements_file:
        for line in requirements_file.readlines():
            line = line.strip('\n')
            if is_a_requirement(line):
                requirements[section].append(line)

with open('README.md', 'r', encoding='utf-8') as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name='ggml',
    version='0.0.1',
    python_requires='>={}.{}'.format(*PYTHON_REQUIRED),
    author='Anton Dmitriev',
    author_email='dmitrievanthony@gmail.com',
    description='GridGain ML Python API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=(
        'https://github.com/apache/ignite/tree/master'
        '/modules/ml/python'
    ),
    packages=setuptools.find_packages(),
    install_requires=requirements['install'],
    tests_require=requirements['tests'],
    setup_requires=requirements['setup'],
    extras_require={
        'docs': requirements['docs'],
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
)
